import sqlite3
import logging
from . import generic


logger = logging.getLogger(__name__)


class Database (generic.Database):
    """Connection to a SQLite database.

    Args:
        name: Name of the database, passed to :py:func:`sqlite3.connect`. The default value "" creates a new in-memory database.
"""
    
    def __init__(self, name=""):
        conn = sqlite3.connect(name, check_same_thread=False)
        super().__init__(conn)
        self.name = name
        self.default_column_type=""
        self.value_placeholder = "?"
        self.temporary_prefix = "temp.temp_"        
    
    def _drop(self, statement):
        logger.debug(f"[{self!r}] Scheduling drop {statement!r}")        
        self._gc_statements.append(statement)
        self._garbage_collect()        
        
    def drop_table(self, table_name):
        """Drop a table from the database. To work around locking issues, the table is first 
        renamed and then dropped once all active iterators on the connection have ended.
        
        Args:
            table_name: The name of the table to drop.
        """        
        quoted_table_name = self.quote_name(table_name)
        temp_name = self._generate_temp_name(prefix="_drop_")
        with self._conn:
            self.execute(f"alter table {quoted_table_name} rename to {temp_name}")
        self._drop(f"drop table {temp_name}")        
        
    def create_function(self, name, nargs, fn):
        """Register a SQLite user-defined function
        
        Args:
            name (str): name of the function in SQLite
            nargs (int): number of arguments
            fn (callable): Python function object
        """
        return self._conn.create_function(name, nargs, fn)
        


from collections import namedtuple, deque
import weakref
import re
import logging

from .sqltables import Table, RowIterator

logger = logging.getLogger(__name__)


# We are forced to implement a garbage collection mechanism in Database since
# "Attempting to DROP a table gets an SQLITE_LOCKED error if there are any active statements
# belonging to the same database connection"
# https://sqlite.org/forum/forumpost/433d2fdb07?raw
class Database:
    """Connection to a Generic SQL database.

    Args:
        conn: DB-API 2.0 connection object

    """
    
    def __init__(self, conn):
        self._conn = conn
        self._next_temp_id = 0
        self._active_iterators = weakref.WeakSet()
        self._gc_statements = deque()
        self.default_column_type = "text"
        self.value_placeholder = "?"
        self.temporary_prefix = "temp_"
        self.temporary_modifier = "temporary"

    def _generate_temp_name(self, prefix=None):
        if prefix is None:
            prefix = self.temporary_prefix
        name = f"{prefix}{self._next_temp_id}"
        self._next_temp_id += 1
        return name

    def quote_name(self, name):
        return ".".join(['"' + part.replace('"', '""') + '"' for part in name.split(".")])

    def _cursor(self, cursor_type):
        return self._conn.cursor()

    def _execute_query(self, *args, **kwargs):
        return self._execute(*args, **kwargs)

    def _execute(self, statement, parameters=None, cursor_type="client"):
        with self._conn:
            cursor = self._cursor(cursor_type)
            try:
                if parameters is not None:
                    logger.debug(f"[{self!r}] Executing {statement!r} with parameters {parameters!r}")
                    cursor.execute(statement, parameters)
                else:
                    logger.debug(f"[{self!r}] Executing {statement!r}")
                    cursor.execute(statement)
            except:
                cursor.close()
                raise
            return cursor

    def _iterate(self, table):
        statement = f"select * from {table.name}"
        result_iterator = RowIterator(statement, table)
        self._active_iterators.add(result_iterator)
        weakref.finalize(result_iterator, self._garbage_collect)
        return result_iterator

    def _drop(self, statement):
        self.execute(statement)

    def _garbage_collect(self):
        self._active_iterators = weakref.WeakSet(x for x in self._active_iterators if x.active)
        if not self._active_iterators and self._gc_statements:
            logger.debug(f"[{self!r}] Starting GC on database {self.name!r} {self!r}")            
            while self._gc_statements:
                statement = self._gc_statements.popleft()
                self.execute(statement)
    
    def query(self, select_stmt, kind="view", parameters=None, bindings={}):
        """Execute an SQL select statement.

        Args:
            select_stmt (str): The SQL select statement to execute. 
                Does not support "with" clauses.
            kind (str): The underlying temporary object to create. Either "view" 
                or "table".
            parameters (list or dict): Query parameters for the SQL statement.
                Only supported if kind is "table"
            bindings (dict(str, Table)): For each key name, make the table available within
                the query as name.
            
        Returns:
            Table: A Table object that represents the result of the query
            
            
        """
        if re.match(r"\s*with\b", select_stmt):
            raise ValueError("sqltables: with clause not supported in query, please use the bindings parameter instead")
        preamble = []
        with_clauses = [
            f"{name} as (select * from {table.name})"
            for name, table in bindings.items() if table.name is not None
        ]
        with_stmt = "with " + ", ".join(with_clauses) if with_clauses else ""
        result_name = self._generate_temp_name()
        statement = f"create {self.temporary_modifier} {kind} {result_name} as {with_stmt} {select_stmt}"
        self.execute(statement, parameters)
        result = Table(name=result_name, db=self)
        result.bindings = bindings
        weakref.finalize(result, self._drop, f"drop {kind} {result_name}")
        return result

    def execute(self, statement, parameters=None):
        cur = self._execute(statement, parameters)
        cur.close()

    def create_table(self, rows=None, column_names=[], column_types={}, name=None):
        """Create a new table.
        
        Args:
            values (iterable(sequence)): The values to insert into the new table, as
                an iterable of rows.
            column_names (list(str)): The column names of the new table.
            column_types (dict(str, str)): Data types for the columns. Note that SQLite mostly uses 
                column types for documentation purposes.
            name (str): The name of the table inside the database. The default value 
                `None` causes a name to be automatically generated.
                
        Returns:
            Table: A new Table object that can be used to manipulate the created table.
        """
        is_temporary = True
        temporary = self.temporary_modifier
        if name is None:
            name = self._generate_temp_name()
        else:
            temporary = ""
            is_temporary = False
        quoted_column_names = ['"' + n.replace('"', '""') + '"' 
                               for n in column_names]
        column_spec = ",".join([f'{q} {column_types.get(x, self.default_column_type)}' for q,x in zip(quoted_column_names, column_names)])
        with self._conn:
            self.execute(f"create {temporary} table {name} ({column_spec})")
            result = Table(name=name, db=self)
            if rows is not None:
                self._insert_values(result, rows, column_names=column_names)
        if is_temporary:
            weakref.finalize(result, self._drop, f"drop table {name}")
        return result

    def open_table(self, name):
        return Table(name=name, db=self)
    
    # Deprecated, use create_table instead
    def load_values(self, values, *, column_names, name=None):
        return self.create_table(rows=values, column_names=column_names, name=name)
    
    def drop_table(self, table_name):
        """Drop a table from the database.
        
        Args:
            table_name: The name of the table to drop.
        """
        quoted_name = self.quote_name(table_name)
        self.execute(f"drop table {quoted_name}")

    def close(self):
        self._conn.close()
        
    def _insert_values(self, table, values, column_names=None):
        if column_names is None:
            it = iter(table)
            column_names = it.column_names
            it.close()
        name = self.quote_name(table.name)
        with self._conn:
            cursor = self._conn.cursor()
            try:
                if isinstance(values, Table):
                    values_name = self.quote_name(values.name)
                    cursor.execute(f"insert into {name} select * from {values_name}")
                else:
                    value_spec = ",".join(self.value_placeholder for _ in column_names)            
                    cursor.executemany(f"insert into {name} values ({value_spec})", values)
            finally:
                cursor.close()



import io
import psycopg2
from . import generic

class Database (generic.Database):
    """Connection to a PostgreSQL database.

    Args:
        name: Name of the database, passed to :py:func:`psycopg2.connect`.

    """
    
    def __init__(self, name):
        conn = psycopg2.connect(name)
        super().__init__(conn)
        self.name = name
        self.value_placeholder = "%s"  

    def _cursor(self, cursor_type):
        if cursor_type == "client":
            return self._conn.cursor()
        elif cursor_type == "server":
            cursor_name = self._generate_temp_name(prefix="cursor")
            return self._conn.cursor(name=cursor_name, withhold=True)
        else:
            raise ValueError(f"Invalid cursor type: {cursor_type!r}")

    def _execute_query(self, statement, parameters=None, cursor_type="client"):
        cur = self._execute(statement, parameters, cursor_type=cursor_type)
        if cursor_type == "server":
            cur.fetchmany(0) # Force PostgreSQL to populate cur.description
        return cur

    def _insert_values(self, table, values, column_names=None):
        def quote_copy_text(value):
            if value is None:
                return r"\N"
            return (str(value)
                    .replace("\\", "\\\\")
                    .replace("\t", r"\t")
                    .replace("\n", r"\n")
                    .replace("\r", r"\r"))

        def encode_rows(rows):
            buf = io.StringIO()
            for row in rows:
                buf.write("\t".join(quote_copy_text(x) for x in row))
                buf.write("\n")
            buf.seek(0)
            return buf
        
        buf = encode_rows(values)
        quoted_name = self.quote_name(table.name)
        with self._conn:
            cursor = self._conn.cursor()
            cursor.copy_from(buf, quoted_name)





import sqlite3
from collections import namedtuple, deque
from io import StringIO
import weakref
import re
import logging


logger = logging.getLogger(__name__)


class RowIterator:
    """An iterator over the rows in a view or table. 
    Never instantiate this directly, created by iterating over a Table object.
    
    Attributes:
        column_names (list(str)): The names of the columns in the table or view.
        Row (class): The :py:class:`collections.namedtuple` class used for representing the rows.
        
    """
    def __init__(self, statement, table):
        self.statement = statement
        self.active = True
        self.table = table
        cur = table.db._execute_query(statement, cursor_type="server")
        self._cur = cur
        self.column_names = [x[0] for x in cur.description]
        self.Row = namedtuple("Row", self.column_names, rename=True)

    def __iter__(self):
        return self
        
    def __next__(self):
        if not self.active:
            raise StopIteration        
        try:
            return self.Row._make(next(self._cur))
        except StopIteration:
            self.close()
            raise

    def close(self):
        self._cur.close()
        self.active = False
        self.table.db._garbage_collect()
        
    def __del__(self):
        if hasattr(self, "_cur"):
            self._cur.close()
        else:
            logger.debug(f"In __del__ of {self!r}:{self.statement}: _cur attribute uninitialized")

class Table:
    """Represents a table or view. 
    Returned by :py:meth:`Database.query`, :py:meth:`Table.view` or :py:meth:`Table.table`. Not to be instantiated directly.
    
    """
    def __init__(self, name, db):
        self.name = name
        self.db = db
        self.bindings = None

    def view(self, select_stmt, *, bindings={}):
        """Create a new view by running a SQL select statement.
        
        Args:
            select_stmt (str): SQL select statement. The special table name `_` 
                (underscore) represents the table associated with `self`.
            bindings (dict(str, Table)): Additional tables to be made accessible
                in the SQL statement.
                
        Returns:
            Table: A new table object representing the result of the query.
                
        """
        return self.db.query(select_stmt, kind="view", bindings=dict(_=self, **bindings))

    def table(self, select_stmt=None, parameters=None, *, bindings={}):
        """Create a new view by running a SQL select statement.
        
        Args:
            select_stmt (str): SQL select statement. The special table name `_` 
                (underscore) represents the table associated with `self`. If `None`,
                defaults to `select * from _`.
            parameters (list or dict): Values for SQL query parameters
            bindings (dict(str, Table)): Additional tables to be made accessible
                in the SQL statement.
                
        Returns:
            Table: A new table object representing the result of the query.

        """

        if select_stmt is None:
            select_stmt = "select * from _"
        return self.db.query(select_stmt, kind="table", parameters=parameters, bindings=dict(_=self, **bindings))
    
    def insert(self, values):
        self.db._insert_values(self, values)

    def __iter__(self):
        """Iterate over the rows from this table.
        
        Returns:
            RowIterator: An iterator over the rows in this table.
            
        """
        return self.db._iterate(self)

    def _repr_markdown_(self, limit=16):
        ascii_punctuation = r"""!"#$%&'()*+,-./:;<=>?@[\]^_`{|}~"""
        def q(x):
            return "".join("\\" + c if c in ascii_punctuation else c for c in x)
        out = StringIO()
        it = iter(self.view(f"select * from _ limit {limit+1}"))
        out.write("|" + "|".join(q(x) for x in it.column_names) + "|\n")
        out.write("|" + "|".join("-" for _ in it.column_names) + "|\n")
        for i,row in enumerate(it):
            if i < limit:
                data = [q(f"{x!r}") for x in row]
            else:
                data = ["..." for _ in row]
            out.write("|" + "|".join(data) + "|\n")
        return out.getvalue()
